/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  Cache fixes for V36 (A3000 Beta ROM) and some code optimizations by Janos
 *  Farkas
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 */


	.global	_LoaderVectorStart, _LoaderVectorEnd, _AltDeviceName, _AltDeviceUnit

    /*
     *	Alert Codes
     */

AT_DeadEnd		= 0x80000000
AN_LILO			= AT_DeadEnd+0x4c000000 | Lilo Alert Base CC000000
AG_NoMemory		= 0x00010000
AG_OpenDev		= 0x00040000
AG_IOError		= 0x00060000
AG_NoSignal		= 0x00070000

    /*
     *	Memory Allocation
     */

MEMF_PUBLIC		= 1
MEMF_CLEAR		= 0x10000
MEMF_PUB_CLR		= MEMF_PUBLIC+MEMF_CLEAR

    /*
     *	Exec I/O
     */

IO_COMMAND		= 28
IO_FLAGS		= 30
IO_ERROR		= 31
IO_LENGTH		= 36
IO_DATA			= 40
IO_OFFSET		= 44
CMD_READ		= 2
IOF_QUICK		= 1

    /*
     *  SCSI
     */

HD_SCSICMD		= 28

SCSI_DATA		= 0
SCSI_LENGTH		= 4
SCSI_COMMAND		= 12
SCSI_CMDLENGTH		= 16
SCSI_FLAGS		= 20
SCSI_STATUS		= 21
SCSI_SENSEDATA		= 22
SCSI_SENSELENGTH	= 26
SCSICMD_SIZEOF		= 30

SCSIF_READ		= 1
SCSIF_AUTOSENSE		= 2

    /*
     *	Trackdisk
     */

TD_MOTOR		= 9
IOEXTTD_SIZE		= 56

    /*
     *	Library
     */

LIB_VERSION		= 20

    /*
     *	Resident Tag
     */

RT_INIT			= 22

    /*
     *	Expansion Library
     */

eb_Flags		= 34
EBB_SILENTSTART		= 6

    /*
     *	Exec Library Vector Offsets
     */

LVOSuperVisor		= -0x1e
LVOFindResident		= -0x60
LVOAlert		= -0x6c
LVOCloseLibrary		= -0x19e
LVOOpenDevice		= -0x1bc
LVOCloseDevice		= -0x1c2
LVODoIO			= -0x1c8
LVOOpenLibrary		= -0x228
LVOCopyMem		= -0x270
LVOCacheClearU		= -0x27c		| Only V37+
LVOCreateIORequest	= -0x28e
LVODeleteIORequest	= -0x294
LVOCreateMsgPort	= -0x29a
LVODeleteMsgPort	= -0x2a0
LVOAllocVec		= -0x2ac
LVOFreeVec		= -0x2b2

    /*
     *	Disk Block Size
     */

LILO_BLOCKSIZE		= 1024
LILO_BLOCKSIZESHIFT	= 10

    /*
     *	68030 Cache Flags
     */

FLUSH_I_AND_D		= 0x00000808

    /*
     *	Lilo Data Offsets (Stack Frame)
     */

SysBase			=   -1*4		| ExecBase
BootRequest		=   -2*4		| Passed by Bootstrap
FuncTable		=   -9*4		| Function Table
FileVector		=  -10*4
FileOffset		=  -11*4
LoaderMem		=  -12*4
AltPort			=  -13*4
BASE_SIZE		=   13*4


.text

    /*
     *	Boot Code
     *
     *	    in:	    a1 = bootrequest
     *		    a6 = execbase
     *
     *	    out:    d0 = error
     */

BootCodeStart:
	movem.l	%d2-%d7/%a2-%a6,-(%sp)

    /*
     *	Set Up the Stack Frame
     */

	link	%a5,#-BASE_SIZE
	lea	BaseVar(%pc),%a0
	move.l	%a5,(%a0)
	move.l	%a6,SysBase(%a5)
	move.l	%a1,BootRequest(%a5)
	move.l	%a1,%a2

    /*
     *	Alternate Boot Device Initialization
     */

	tst.b	AltDeviceName(%pc)
	jeq	NoAltDevice
	jsr	LVOCreateMsgPort(%a6)
	move.l	%d0,AltPort(%a5)
	jne	AltPortOK
	move.l	#AN_LILO+AG_NoSignal,%d7
	jmp	LVOAlert(%a6)
AltPortOK:
	move.l	%d0,%a0
	moveq	#IOEXTTD_SIZE,%d0
	jsr	LVOCreateIORequest(%a6)
	tst.l	%d0
	jne	1f
	move.l	#AN_LILO+AG_NoMemory,%d7
	jmp	LVOAlert(%a6)
1:	move.l	%d0,%a2
	move.l	%d0,%a1
	lea	AltDeviceName(%pc),%a0
	move.l	AltDeviceUnit(%pc),%d0
	clr.l	%d1
	jsr	LVOOpenDevice(%a6)
	tst.l	%d0
	jeq	NoAltDevice
	move.l	#AN_LILO+AG_OpenDev,%d7
	jmp	LVOAlert(%a6)

    /*
     *	Load the Boot Loader
     */

NoAltDevice:
	lea	LoaderVectorStart(%pc),%a3
	move.l	(%a3)+,%d2		| file size
	addq.l	#4,%a3			| skip count
	move.l	(%a3)+,%d3		| sector shift
	moveq	#1,%d0
	lsl.l	%d3,%d0
	add.l	%d2,%d0
	add.l	#8192,%d0
	move.l	#MEMF_PUBLIC,%d1
	jsr	LVOAllocVec(%a6)
	move.l	%d0,LoaderMem(%a5)
	jne	1f
	move.l	#AN_LILO+AG_NoMemory,%d7
	jmp	LVOAlert(%a6)
1:	move.l	%d0,%a4
	moveq	#~0,%d4			| max 4GB sector
	lsr.l	%d3,%d4			
loop:
	move.l	4(%a3),%d5		| sector count
	move.l	%d5,%d1
	lsl.l	%d3,%d5
	move.l	(%a3),%d0		| start sector
	jeq	clear
	move.l	%d0,%a0
	add.l	%d1,%a0
	cmp.l	%d4,%a0
	bcc	scsi
	move.w	#CMD_READ,IO_COMMAND(%a2)
	move.l	%a4,IO_DATA(%a2)
	move.l	%d5,IO_LENGTH(%a2)
	lsl.l	%d3,%d0
	move.l	%d0,IO_OFFSET(%a2)
	jra	doio
scsi:
	move.w	#HD_SCSICMD,IO_COMMAND(%a2)
	lea	ScsiCmd(%pc),%a0
	move.l	%a0,IO_DATA(%a2)
	move.l  #SCSICMD_SIZEOF,IO_LENGTH(%a2)
	clr.l	IO_OFFSET(%a2)
	move.l	%a4,SCSI_DATA(%a0)
	move.l	%d5,SCSI_LENGTH(%a0)
	move.w	#10,SCSI_CMDLENGTH(%a0)
	move.b	#SCSIF_AUTOSENSE+SCSIF_READ,SCSI_FLAGS(%a0)
	lea	SenseData(%pc),%a1
	move.l  %a1,SCSI_SENSEDATA(%a0)
	move.w	#4,SCSI_SENSELENGTH(%a0)
	lea	ReadCmd(%pc),%a1
	move.l	%a1,SCSI_COMMAND(%a0)
	move.l	%d0,2(%a1)
	move.w	%d1,7(%a1)
doio:
	move.l	%a2,%a1
	jsr	LVODoIO(%a6)
	tst.b	%d0
	jeq	next
	move.l	#AN_LILO+AG_IOError,%d7
	lsl.l	#8,%d0
	or.w	%d0,%d7
	or.b	IO_ERROR(%a2),%d0
	jmp	LVOAlert(%a6)
clear:
	lea	(%a4,%d0.l),%a1
	move.l	%a4,%a0
0:	clr.l	(%a0)+
	cmp.l	%a0,%a1
	jhi	0b
next:
	addq.l	#8,%a3
	add.l	%d5,%a4
	sub.l	%d5,%d2
	jhi	loop

    /*
     *	Stop the floppy motor
     */

	move.l	%a2,%a1
	move.w	#TD_MOTOR,IO_COMMAND(%a1)
	move.b	#IOF_QUICK,IO_FLAGS(%a1)
	clr.l	IO_LENGTH(%a1)
	jsr	LVODoIO(%a6)

    /*
     *	Alternate Boot Device Clean Up
     */

	tst.b	AltDeviceName(%pc)
	jeq	NoAltDevice2
	move.l	%a2,%a1
	jsr	LVOCloseDevice(%a6)
	move.l	%a2,%a0
	jsr	LVODeleteIORequest(%a6)
	move.l	AltPort(%a5),%a0
	jsr	LVODeleteMsgPort(%a6)

NoAltDevice2:
    /*
     *	Execute the Boot Loader
     */

	move.l	LoaderMem(%a5),%a2
	lea	LVOCacheClearU(%a6),%a1
	cmp.w	#37,LIB_VERSION(%a6)
	jcc	Cache37			| Only V37+
	lea	Cache36(%pc),%a1	| Home-made Cache Flush
Cache37:
	jsr	(%a1)			| Do the Flush
	move.l	BootRequest(%a5),-(%sp)	| Pass Arg
	move.l	%a2,%a0
	jsr	([%a0],%a0)
	addq.l	#4,%sp
	move.l	%d0,%d2

    /*
     *  Free the Boot Loader
     */

	move.l	%a2,%a1
	jsr	LVOFreeVec(%a6)

	tst.l	%d2
	jne	BootAmigaOS
BootError:
	moveq	#-1,%d0
	jra	BootReturn

    /*
     *	Boot the Standard Amiga Operating System
     */

BootAmigaOS:
	lea	ExpansionName(%pc),%a1
	moveq	#37,%d0
	jsr	LVOOpenLibrary(%a6)
	tst.l	%d0
	jeq	NoExpansion37
	move.l	%d0,%a1
	bset	#EBB_SILENTSTART,eb_Flags(%a1)
	jsr	LVOCloseLibrary(%a6)
NoExpansion37:
	lea	DOSName(%pc),%a1
	jsr	LVOFindResident(%a6)
	tst.l	%d0
	jeq	BootError
	move.l	%d0,%a0
	move.l	RT_INIT(%a0),%a0
	moveq	#0,%d0
BootReturn:
	unlk	%a5
	movem.l	(%sp)+,%d2-%d7/%a2-%a6
	rts

    /*
     *	This code only runs on pre-2.0 (also early A3000 Boot ROMs)
     *	I assume there are still a lot of these (I have one too 8)
     *
     *	No need to check for 040, because this ROM doesn't work at all
     *	on anything but a 030...
     */

Cache36:
	move.l	%a5,%a4
	lea	CacheSV(%pc),%a5
	jsr	LVOSuperVisor(%a6)
	move.l	%a4,%a4
	rts
CacheSV:
	movec	%cacr,%d0
	or.w	#FLUSH_I_AND_D,%d0	| clear 030 caches
	movec	%d0,%cacr
	rte



    /*
     *	Strings
     */

ExpansionName:
	.asciz	"expansion.library"
DOSName:
	.asciz	"dos.library"


    /*
     *	File Data for the Loader Program
     */

	.align	4
_LoaderVectorStart:
LoaderVectorStart:
	.int	0		| File Size
	.fill	10,4		| Max. 68 Data Blocks
_LoaderVectorEnd:


    /*
     *	Alternate Boot Device
     */

_AltDeviceName:
AltDeviceName:
	.fill	32		| Max. 32 bytes (hardcoded!)

_AltDeviceUnit:
AltDeviceUnit:
	.int	0

ScsiCmd:
	.fill	SCSICMD_SIZEOF,1
ReadCmd:
	.byte	0x28
	.fill	9,1
SenseData:
	.int	0

    /*
     *	Base Pointer for Local Variables
     */

BaseVar:
	.ascii	"END!"		| Will be overwritten!
