/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 */


    /*
     *	Function Table (Passed at Startup)
     */

#ifndef __ASSEMBLY__

struct LiloTable {
    void *lilo_open;
    void *lilo_close;
    void *lilo_read;
    void *lilo_seek;
    void *lilo_load;
    void *lilo_unload;
    void *lilo_exec;
    struct IOStdReq *boot_request;
};

extern const struct LiloTable *LiloTable;


    /*
     *	Open a File
     */

static inline void Lilo_Open(const struct FileVector *vector)
{
	register void **a1 asm("a1") = LiloTable->lilo_open;
	register const struct FileVector *a0 asm("a0") = vector;

	asm volatile ("jsr %%a1@"
		      : "=r" (a0), "=r" (a1)
		      : "0" (a0), "1" (a1)
		      : "d0", "d1", "memory");
}


    /*
     *	Close the File
     */

static inline void Lilo_Close(void)
{
	register void **a1 asm("a1") = LiloTable->lilo_close;

	asm volatile ("jsr %%a1@"
		      : "=r" (a1)
		      : "0" (a1)
		      : "a0", "d0", "d1", "memory");
}


    /*
     *	Read a Block from the File
     */

static inline void Lilo_Read(void *dest, u_long length)
{
	register void **a1 asm("a1") = LiloTable->lilo_read;
	register void *a0 asm("a0") = dest;
	register u_long d0 asm("d0") = length;

	asm volatile ("jsr %%a1@"
		      : "=r" (a0), "=r" (a1), "=r" (d0)
		      : "0" (a0), "1" (a1), "2" (d0)
		      : "d1", "memory");
}


    /*
     *	Absolute Seek
     */

static inline void Lilo_Seek(u_long offset)
{
	register void **a1 asm("a1") = LiloTable->lilo_seek;
	register u_long d0 asm("d0") = offset;

	asm volatile ("jsr %%a1@"
		      : "=r" (a1), "=r" (d0)
		      : "0" (a1), "1" (d0)
		      : "a0", "d1", "memory");
}


    /*
     *	Load a File
     */

static inline u_long *Lilo_Load(const u_long *vector)
{
	register u_long *_res asm("d0");
	register void **a1 asm("a1") = LiloTable->lilo_load;
	register const u_long *a0 asm("a0") = vector;

	asm volatile ("jsr %%a1@"
		      : "=r" (_res), "=r" (a0), "=r" (a1)
		      : "1" (a0), "2" (a1)
		      : "d1", "memory");
	return(_res);
}


    /*
     *	Unload a File
     */

static inline void Lilo_UnLoad(u_long *data)
{
	register void **a1 asm("a1") = LiloTable->lilo_unload;
	register u_long *a0 asm("a0") = data;

	asm volatile ("jsr %%a1@"
		      : "=r" (a0), "=r" (a1)
		      : "0" (a0), "1" (a1)
		      : "d0", "d1", "memory");
}


    /*
     *	Execute a Loaded File
     */

static inline long Lilo_Exec(u_long *data, long arg)
{
	register long _res asm("d0");
	register void **a1 asm("a1") = LiloTable->lilo_exec;
	register u_long *a0 asm("a0") = data;
	register long d0 asm("d0") = arg;

	asm volatile ("jsr %%a1@"
		      : "=r" (_res), "=r" (a0), "=r" (a1)
		      : "0" (d0), "1" (a0), "2" (a1)
		      : "d1", "memory");
	return(_res);
}

#else /* __ASSEMBLY__ */

Lilo_Open	= 0
Lilo_Close	= 4
Lilo_Read	= 8
Lilo_Seek	= 12
Lilo_Load	= 16
Lilo_Unload	= 20
Lilo_Exec	= 24

#endif /* __ASSEMBLY__ */
