/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 */


#define V36_COMPATIBILITY	/* Make it work with the A3000 Beta ROM */

#include <stddef.h>
#include <stdarg.h>
#include <sys/types.h>

#include "amigahw.h"

#include "lilo.h"
#include "bootlib.h"
#include "loader_proto.h"
#include "linuxboot.h"
#include "tools.h"


    /*
     *  Program Entry
     *
     *  Pass the Pointer to the LiloTable (in d0) to Main
     */


asm(".text\n"
"	.globl	_start\n"
"	.align	4\n"
"_start:\n"
"	move.l	%a5,-(%sp)\n"
"	move.l	%a0,%d0		| start\n"
"	move.l	8(%a0),%d1	| reloc_cnt\n"
"	move.l	4(%a0),%a0	| reloc_data + start\n"
"	add.l	%d0,%a0\n"
"1:	move.l	(%a0)+,%a1	| *(*reloc_data++ + start) += start\n"
"	add.l	%d0,(%a1,%d0)\n"
"	subq.l	#1,%d1		| while (--reloc_cnt > 0)\n"
"	jcc	1b\n"
"\n"
"	move.l	4:w,%a6\n"
"	lea	Install060(%pc),%a5\n"
"	jsr	-0x1e(%a6)	| SuperVisor\n"
"\n"
"	lea     -0x27c(%a6),%a1	| CacheClearU\n"
"	cmp.w	#37,20(%a6)	| LIB_VERSION\n"
"	jcc	1f\n"
"	lea	Cache36(%pc),%a1\n"
"1:	jsr	%a1@\n"
"\n"
"	lea	__bss_start,%a0\n"
"	lea	__end,%a1\n"
"	jra	2f\n"
"1:	clr.l	(%a0)+\n"
"2:	cmp.l	%a0,%a1\n"
"	jhi	1b\n"
"\n"
"	move.l	(%sp)+,%a5\n"
"	jmp	Main\n"
"\n"
"Cache36:\n"
"	lea	CacheSV(%pc),%a5\n"
"	jmp	-0x1e(%a6)	| SuperVisor\n"
"CacheSV:\n"
"	movec	%cacr,%d0\n"
"	orw	#0x0808,%d0\n"
"	movec	%d0,%cacr\n"
"	rte\n"
"\n"
"Install060:\n"
"	movec	%vbr,%a0\n"
"	lea	_060_isp_unimp(%pc),%a1\n"
"	move.l	%a1,0xf4(%a0)\n"
"	rte\n"
);


const char LiloVersion[] = VERSION;


    /*
     *	Lilo Alert
     */

#define AN_LILO		(AT_DeadEnd|0x4c000000)	/* `L' */
#define AO_LiloMap	(0x004d4150)		/* `MAP': Corrupt Map File */


    /*
     *	Console Echo Modes
     */

#define ECHO_NONE	(0)
#define ECHO_NORMAL	(1)
#define ECHO_DOT	(2)


    /*
     *	Boot Library Functions
     */

static struct IOStdReq *BootRequest;


    /*
     *	Library Bases
     */

#ifdef V36_COMPATIBILITY
#define LIB_VERS	(36)
#else /* V36_COMPATIBILITY */
#define LIB_VERS	(37)
#endif /* V36_COMPATIBILITY */

const struct ExecBase *SysBase = NULL;
const struct ExpansionBase *ExpansionBase = NULL;
const struct GfxBase *GfxBase = NULL;
static const struct IntuitionBase *IntuitionBase = NULL;
static const struct InputBase *InputBase = NULL;
#ifdef V36_COMPATIBILITY
static const struct KeyboardBase *KeyboardBase = NULL;
#endif /* V36_COMPATIBILITY */


    /*
     *	Console Stuff
     */

static u_char Title[] = "Amiga Linux/m68k Loader";

static struct NewScreen NewScreen = {
    0, 0, 0, 0, 2, 0, 1, HIRES, CUSTOMSCREEN, NULL, NULL, NULL, NULL
};
static struct NewWindow NewWindow = {
    0, 0, 0, 0, 2, 3, 0, WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE, NULL, NULL,
    NULL, NULL, NULL, 0, 0, 0, 0, CUSTOMSCREEN
};
static struct Screen *Screen;
static struct Window *Window;
static struct MsgPort *ConsolePort;
static struct IOStdReq *ConsoleRequest = NULL;
static struct MsgPort *TimerPort;
static struct timerequest *TimerRequest;
static struct MsgPort *InputPort;
static struct IOStdReq *InputRequest;
#ifdef V36_COMPATIBILITY
static struct MsgPort *KeyboardPort;
static struct IOStdReq *KeyboardRequest;
static u_char KeyboardMatrix[13];
#endif /* V36_COMPATIBILITY */
static u_long TimeOutSignalMask = 0;
static char FmtBuffer[256], InBuffer[CL_SIZE];
static u_long ConsoleOpened = FALSE;


    /*
     *	Map Data
     */

#define MAXVECTORSIZE	(64)		/* Should be sufficient */

struct FileExtentData {
    u_int32_t MaxVectorSize;
    struct FileExtent Extent;
    struct Extent Data[MAXVECTORSIZE];
} _MapExtentData = {
    MAXVECTORSIZE
};

char _MapDeviceName[MAXALTDEVSIZE] = "\0\n";
struct DeviceDef _MapDevice = {
    .device = _MapDeviceName,
};

static u_char *MapData;
static u_long MapSize = 0, MapOffset = 0;
static struct FileDef MapFile = {
    .device = &_MapDevice,
    .Extent = &_MapExtentData.Extent,
};
static struct BootOptions *BootOptions;
static struct BootRecord *BootRecords;
static struct FileDef *Files;
static struct DeviceDef *Devices;


static u_short MasterMode = FALSE;
static u_short Auto = FALSE;
static u_short Aux = FALSE;
static u_short Debug = FALSE;
static const char *Prompt;
static u_short MinKick = 0;
static u_short MaxKick = 65535;


    /*
     *	Linux/m68k Bootstrap Stuff
     */

static const struct BootRecord *DefaultBootRecord;
struct BootData BootData;


    /*
     *	Function Prototypes
     */

u_long Main(struct IOStdReq *boot_request);
static u_long GetDebugTag(void);
static const char *GetTagName(u_long tag);
static const struct TagRecord *GetNextTag(void);
static void UnexpectedTag(const struct TagRecord *tr);
static int EqualStrings(const char *s1, const char *s2);
static void ReadMapData(void);
static void FreeMapData(void);
static const struct FileDef *FindFile(const char *path);
#ifdef V36_COMPATIBILITY
static u_short PeekQualifier36(void);
#endif /* V36_COMPATIBILITY */
static u_long ProbeConsole(void);
static void OpenConsole(void);
static void CloseConsole();
static int SerialGetChar(void);
static void SerialPutChar(char c);
static long GetChar(void);
static long ReadLine(u_long mode);
static void SetTimeOut(u_long seconds);
static int TimedOut(void);
static void ClearTimeOut(void);
static void Sleep(u_long micros);
static int IsAvailable(const struct BootRecord *record);
static void GetBootOS(void);
static u_long BootLinux(const struct BootOptions *global,
			const struct BootData *boot, int ppc);
static u_long BootNetBSD(struct BootData *boot);
static int Open(const char *path);
static int OpenFile(const struct FileDef *file);
static int Seek(int fd, int offset);
static int Read(int fd, char *buf, int count);
static void Close(int fd);
static int FileSize(const char *path);


    /*
     *	Main Loader Routine
     *
     *	This routine should return FALSE in case of an error (which will cause
     *	a reboot), and TRUE if the user wants to boot AmigaOS.
     *	In all other cases it won't return :-)
     */

u_long Main(struct IOStdReq *boot_request)
{
    u_long res = FALSE;
    u_short interactive = FALSE;

    //asm("move.w #'a'+256,0xdff030; 9: btst #5,0xdff018; jeq 9b");
    /* Boot Library */
    BootRequest = boot_request;

    /* Open Libraries */
    SysBase = *(struct ExecBase **)4;
    ExpansionBase = (struct ExpansionBase *)OpenLibrary("expansion.library", LIB_VERS);
    if (!ExpansionBase)
	Alert(AN_LILO|AG_OpenLib|AO_ExpansionLib);
    GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", LIB_VERS);
    if (!GfxBase)
	Alert(AN_LILO|AG_OpenLib|AO_GraphicsLib);

    /* Open Timer Device */
    TimerPort = CreateMsgPort();
    if (!TimerPort)
	Alert(AN_LILO|AG_NoSignal);
    TimerRequest = CreateIORequest(TimerPort, sizeof(struct timerequest));
    if (!TimerRequest)
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("timer.device", UNIT_VBLANK,
		   (struct IORequest *)TimerRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_TimerDev);

    /* Open Input Device */
    InputPort = CreateMsgPort();
    if (!InputPort)
	Alert(AN_LILO|AG_NoSignal);
    InputRequest = CreateIORequest(InputPort, sizeof(struct IOStdReq));
    if (!InputRequest)
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("input.device", 0, (struct IORequest *)InputRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_Unknown);
    InputBase = (struct InputBase *)InputRequest->io_Device;

    /* Configuration */
    ReadMapData();

    if (!Auto || ProbeConsole()) {
	OpenConsole();
	interactive = TRUE;
    }

    if (SysBase->Version < MinKick || SysBase->Version > MaxKick) {
	Printf("\nWarning: Kickstart version %ld is out of range %ld-%ld\n\n",
	       SysBase->Version, MinKick, MaxKick);
	if (!interactive) {
	    Puts("Autobooting AmigaOS...\n\n");
	    res = TRUE;
	    goto done;
	}
    }

    do {
	DefaultBootRecord = FindBootRecord(NULL);

	if (!DefaultBootRecord) {
	    Puts("\nPanic: No boot Operating System!\n");
	    break;
	}
	FillBootData(DefaultBootRecord, &BootData, TRUE);
	if (interactive)
	    GetBootOS();

	switch (BootData.OSType) {
	    case BOS_AMIGA:
		Puts("\nBooting AmigaOS...\n\n");
		res = TRUE;
		break;
	    case BOS_LINUX:
		res = BootLinux(BootOptions, &BootData, 0);
		break;
	    case BOS_LINUXPPC:
		res = BootLinux(BootOptions, &BootData, 1);
		break;
	    case BOS_NETBSD:
		Puts("\nBooting NetBSD-Amiga...\n\n");
		res = BootNetBSD(&BootData);
		break;
	}
	if (res)
	    break;
	Puts("\nCouldn't boot the desired Operating System\n");
    } while (interactive);

done:
    CloseConsole();
    FreeMapData();

    /* Close Input Device */
    CloseDevice((struct IORequest *)InputRequest);
    DeleteIORequest(InputRequest);
    DeleteMsgPort(InputPort);

    /* Close Timer Device */
    CloseDevice((struct IORequest *)TimerRequest);
    DeleteIORequest(TimerRequest);
    DeleteMsgPort(TimerPort);

    /* Close Libraries */
    CloseLibrary((struct Library *)GfxBase);
    CloseLibrary((struct Library *)ExpansionBase);

    return(res);
}


    /*
     *	Get the Value of the Debug Tag (if it exists)
     */

static u_long GetDebugTag(void)
{
    const struct TagRecord *tr;
    u_long offset;
    int debug = FALSE;

    for (offset = 0; offset + sizeof(struct TagRecord) + sizeof(u_long) <= MapSize;
    	 offset += sizeof(struct TagRecord) + tr->Size) {
	tr = (struct TagRecord *)&MapData[offset];
	if (tr->Tag == TAG_DEBUG) {
	    debug = tr->Data[0];
	    break;
	}
    }
    return(debug);
}


    /*
     *	Get the Name for a Tag
     */

struct TagName {
    u_long Tag;
    const char *Name;
};

static struct TagName TagNames[] = {
    { TAG_LILO, "TAG_LILO" },
    { TAG_EOF, "TAG_EOF" },
    { TAG_HEADER, "TAG_HEADER" },
    { TAG_HEADER_END, "TAG_HEADER_END" },
    { TAG_DEFAULT, "TAG_DEFAULT" },
    { TAG_AUTO, "TAG_AUTO" },
    { TAG_TIME_OUT, "TAG_TIME_OUT" },
    { TAG_AUX, "TAG_AUX" },
    { TAG_BAUD, "TAG_BAUD" },
    { TAG_MASTER_PASSWORD, "TAG_MASTER_PASSWORD" },
    { TAG_DEBUG, "TAG_DEBUG" },
    { TAG_PROMPT, "TAG_PROMPT" },
    { TAG_KICKRANGE, "TAG_KICKRANGE" },
    { TAG_G_CHIP_RAM_SIZE, "TAG_G_CHIP_RAM_SIZE" },
    { TAG_G_FAST_RAM_CHUNK, "TAG_G_FAST_RAM_CHUNK" },
    { TAG_G_MODEL, "TAG_G_MODEL" },
    { TAG_G_PROCESSOR, "TAG_G_PROCESSOR" },
    { TAG_BOOT_RECORD, "TAG_BOOT_RECORD" },
    { TAG_BOOT_RECORD_END, "TAG_BOOT_RECORD_END" },
    { TAG_ALIAS, "TAG_ALIAS" },
    { TAG_OS_TYPE, "TAG_OS_TYPE" },
    { TAG_KERNEL, "TAG_KERNEL" },
    { TAG_ARGUMENTS, "TAG_ARGUMENTS" },
    { TAG_PASSWORD, "TAG_PASSWORD" },
    { TAG_CHIP_RAM_SIZE, "TAG_CHIP_RAM_SIZE" },
    { TAG_FAST_RAM_CHUNK, "TAG_FAST_RAM_CHUNK" },
    { TAG_MODEL, "TAG_MODEL" },
    { TAG_RAMDISK, "TAG_RAMDISK" },
    { TAG_PROCESSOR, "TAG_PROCESSOR" },
    { TAG_FILE_DEF, "TAG_FILE_DEF" },
    { TAG_FILE_DEF_END, "TAG_FILE_DEF_END" },
    { TAG_VECTOR, "TAG_VECTOR" },
    { TAG_FILE_EXTENT, "TAG_FILE_EXTENT" },
    { TAG_FILE_DEVICE, "TAG_FILE_DEVICE" },
    { TAG_DEVICE_DEF, "TAG_DEVICE_DEF" },
    { TAG_DEVICE_DEF_END, "TAG_DEVICE_DEF_END" },
    { TAG_DEVICE_NAME, "TAG_DEVICE_NAME" },
    { TAG_DEVICE_UNIT, "TAG_DEVICE_UNIT" }
};

static const char *GetTagName(u_long tag)
{
    u_int i;

    for (i = 0; i < arraysize(TagNames); i++)
	if (TagNames[i].Tag == tag)
	    return(TagNames[i].Name);
    return(NULL);
}


    /*
     *	Get the Next Tag from the Map Data
     */

static const struct TagRecord *GetNextTag(void)
{
    static const struct TagRecord *tr = NULL;

    if (tr)
	MapOffset += sizeof(struct TagRecord) + tr->Size;
    if (MapOffset + sizeof(struct TagRecord) > MapSize)
	Alert(AN_LILO|AO_LiloMap);
    tr = (struct TagRecord *)&MapData[MapOffset];
    return(tr);
}


    /*
     *	Unexpected Tag
     */

static void UnexpectedTag(const struct TagRecord *tr)
{
    const char *name = GetTagName(tr->Tag);

    if (name)
	Printf("Unexpected tag `%s' at offset 0x%08lx\n", name, MapOffset);
    else
	Printf("Unknown tag 0x%08lx at offset 0x%08lx\n", tr->Tag, MapOffset);
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *	Read the Map Data
     */

static void ReadMapData(void)
{
    const struct TagRecord *tr;
    struct BootRecord *record;
    struct BootRecord **p1 = &BootRecords;
    struct FileDef *file;
    struct FileDef **p2 = &Files;
    struct DeviceDef *device;
    struct DeviceDef **p3 = &Devices;
    u_short isdefault = FALSE;
    int i;

    if (!MapFile.device->device[0])
	MapFile.device = NULL;
    OpenFile(&MapFile);
    MapSize = MapFile.Extent->Size;
    MapData = AllocMem(MapSize, MEMF_PUBLIC|MEMF_CLEAR);
    if (!MapData)
	Alert(AN_LILO|AG_NoMemory);
    Read(1, MapData, MapSize);
    Close(1);
    if ((Debug = GetDebugTag()))
	OpenConsole();

    /* File Identification */
    tr = GetNextTag();
    if (tr->Tag != TAG_LILO)
	Alert(AN_LILO|AO_LiloMap);

    /* Boot Options */
    if (Debug)
	Puts("  [Boot Options]\n");
    tr = GetNextTag();
    if (tr->Tag != TAG_HEADER)
	Alert(AN_LILO|AO_LiloMap);
    BootOptions = AllocMem(sizeof(*BootOptions), MEMF_PUBLIC|MEMF_CLEAR);
    if (!BootOptions)
	Alert(AN_LILO|AG_NoMemory);
    for (tr = GetNextTag(); tr->Tag != TAG_HEADER_END; tr = GetNextTag()) {
	//Printf("GTag:%ld,%lx,%lx\n", tr->Tag, tr->Size, tr->Data[0]);
	switch (tr->Tag) {
	    case TAG_DEFAULT:
		BootOptions->Default = (const char *)tr->Data;
		break;
	    case TAG_AUTO:
		BootOptions->Auto = (const u_long *)tr->Data;
		break;
	    case TAG_TIME_OUT:
		BootOptions->TimeOut = (const u_long *)tr->Data;
		break;
	    case TAG_AUX:
		BootOptions->Aux = (const u_long *)tr->Data;
		break;
	    case TAG_BAUD:
		BootOptions->Baud = (const u_long *)tr->Data;
		break;
	    case TAG_MASTER_PASSWORD:
		BootOptions->MasterPassword = (const char *)tr->Data;
		break;
	    case TAG_DEBUG:
		BootOptions->Debug = (const u_long *)tr->Data;
		break;
	    case TAG_PROMPT:
		BootOptions->Prompt = (const char *)tr->Data;
		break;
	    case TAG_KICKRANGE:
		BootOptions->KickRange = (const u_short *)tr->Data;
		break;
	    case TAG_G_CHIP_RAM_SIZE:
		BootOptions->ChipSize = (const u_long *)tr->Data;
		break;
	    case TAG_G_FAST_RAM_CHUNK:
		for (i = 0; i < arraysize(BootOptions->FastChunks); i++)
		    if (!BootOptions->FastChunks[i])
			break;
		if (i < arraysize(BootOptions->FastChunks))
		    BootOptions->FastChunks[i] = (const u_long *)tr->Data;
		else
		    UnexpectedTag(tr);
		break;
	    case TAG_G_MODEL:
		BootOptions->Model = (const u_long *)tr->Data;
		break;
	    case TAG_G_PROCESSOR:
		BootOptions->Processor = (const u_long *)tr->Data;
		break;
	    default:
		UnexpectedTag(tr);
		break;;
	}
    }

    if (BootOptions->Auto)
	Auto = *BootOptions->Auto;
    if (BootOptions->Aux)
	Aux = *BootOptions->Aux;
    if (BootOptions->Debug)
	Debug = *BootOptions->Debug;
    if (BootOptions->Baud) {
	u_long baud = *BootOptions->Baud;
	if (baud) {
	    ciab.ddra |= (SER_DTR | SER_RTS);
	    ciab.ddra &= ~(SER_DCD | SER_CTS | SER_DSR);
	    ciab.pra &= ~(SER_DTR | SER_RTS);
	    custom.serper = (5 * SysBase->ex_EClockFrequency + baud / 2) / baud - 1;
	    custom.intena = 0x0800;
	}
    }
    if (!BootOptions->Default)
	isdefault = TRUE;
    if (BootOptions->KickRange) {
	MinKick = BootOptions->KickRange[0];
	MaxKick = BootOptions->KickRange[1];
    }
    Prompt = BootOptions->Prompt ? BootOptions->Prompt : "Boot image: ";

    /* Boot Records */
    if (Debug)
	Puts("  [Boot Records]\n");
    tr = GetNextTag();
    while (tr->Tag == TAG_BOOT_RECORD) {
	record = AllocMem(sizeof(*record), MEMF_PUBLIC|MEMF_CLEAR);
	if (!record)
	    Alert(AN_LILO|AG_NoMemory);
	record->Label = (const char *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_BOOT_RECORD_END;
	     tr = GetNextTag()) {
	    //Printf("BTag:%ld,%lx,%lx\n", tr->Tag, tr->Size, tr->Data[0]);
	    switch (tr->Tag) {
		case TAG_ALIAS:
		    record->Alias = (const u_char *)tr->Data;
		    break;
		case TAG_OS_TYPE:
		    record->OSType = (const u_long *)tr->Data;
		    break;
		case TAG_KERNEL:
		    record->Kernel = (const u_char *)tr->Data;
		    break;
		case TAG_ARGUMENTS:
		    record->Args = (const u_char *)tr->Data;
		    break;
		case TAG_PASSWORD:
		    record->Password = (const u_char *)tr->Data;
		    break;
		case TAG_CHIP_RAM_SIZE:
		    record->ChipSize = (const u_long *)tr->Data;
		    break;
		case TAG_FAST_RAM_CHUNK:
		    for (i = 0; i < arraysize(record->FastChunks); i++)
			if (!record->FastChunks[i])
			    break;
		    if (i < arraysize(record->FastChunks))
			record->FastChunks[i] = (const u_long *)tr->Data;
		    else
			UnexpectedTag(tr);
		    break;
		case TAG_MODEL:
		    record->Model = (const u_long *)tr->Data;
		    break;
		case TAG_RAMDISK:
		    record->Ramdisk = (const u_char *)tr->Data;
		    break;
		case TAG_PROCESSOR:
		    record->Processor = (const u_long *)tr->Data;
		    break;
		default:
		    UnexpectedTag(tr);
		    break;;
	    }
	}

	if (EqualStrings(BootOptions->Default, record->Label) ||
	    EqualStrings(BootOptions->Default, record->Alias))
	    isdefault = TRUE;

	if (Debug) {
	    if (isdefault)
		Printf("\e[1m");
	    Printf("    %s\n", record->Label);
	    if (isdefault) {
		Printf("\e[0m");
		isdefault = FALSE;
	    }
	}

	*p1 = record;
	p1 = &record->Next;
	tr = GetNextTag();
    }

    /* Device Definitions */
    if (Debug)
	Puts("  [Device Definitions]\n");
    for (; tr->Tag == TAG_DEVICE_DEF; tr = GetNextTag()) {
	device = AllocMem(sizeof(*device), MEMF_PUBLIC|MEMF_CLEAR);
	if (!device)
	    Alert(AN_LILO|AG_NoMemory);
	device->dev_nr = *(dev_t *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_DEVICE_DEF_END; tr = GetNextTag()) {
	    switch (tr->Tag) {
	    case TAG_DEVICE_NAME:
		device->device = (char *)tr->Data;
		break;
	    case TAG_DEVICE_UNIT:
		device->unit = tr->Data[0];
		break;
	    default:
		UnexpectedTag(tr);
		break;
	    }
	}
	if (Debug) {
	    Printf("    %u:%u (", major(device->dev_nr), minor(device->dev_nr));
	    if (device->device)
		Printf("%s:%u)\n", device->device, device->unit);
	    else
		Puts("not available)\n");
	}

	*p3 = device;
	p3 = &device->Next;
    }

    /* File Definitions */
    if (Debug)
	Puts("  [File Definitions]\n");
    for (; tr->Tag == TAG_FILE_DEF; tr = GetNextTag()) {
	file = AllocMem(sizeof(*file), MEMF_PUBLIC|MEMF_CLEAR);
	if (!file)
	    Alert(AN_LILO|AG_NoMemory);
	file->Path = (const char *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_FILE_DEF_END; tr = GetNextTag()) {
	    switch (tr->Tag) {
	    case TAG_VECTOR:
		file->Vector = (struct FileVector *)tr->Data;
		break;
	    case TAG_FILE_EXTENT:
		file->Extent = (struct FileExtent *)tr->Data;
		break;
	    case TAG_FILE_DEVICE:
		for (device = Devices; device; device = device->Next)
		    if (device->dev_nr == *(dev_t *)tr->Data)
			break;
		file->device = device;
		break;
	    default:
		UnexpectedTag(tr);
		break;
	    }
	}
	if (Debug) {
	    Printf("    %s (", file->Path);
	    if (file->Vector)
		Printf("%ld bytes)\n", file->Vector->Size);
	    else
		Puts("not available)\n");
	}

	*p2 = file;
	p2 = &file->Next;
    }

    if (tr->Tag != TAG_EOF)
	UnexpectedTag(tr);
}


    /*
     *	Free the Map Data
     */

static void FreeMapData(void)
{
    struct BootRecord *record, *next;

    for (record = BootRecords; record; record = next) {
	next = record->Next;
	FreeMem(record, sizeof(struct BootRecord));
    }
    if (BootOptions)
	FreeMem(BootOptions, sizeof(struct BootOptions));
    if (MapData)
	FreeMem(MapData, MapSize);
}


    /*
     *	Find a Boot Record by Name
     */

const struct BootRecord *FindBootRecord(const char *name)
{
    const struct BootRecord *record;

    if (!name)
	name = BootOptions->Default;
    for (record = BootRecords; record; record = record->Next) {
	if (EqualStrings(name, record->Label) ||
	    EqualStrings(name, record->Alias))
	    break;
    }
    return(record);
}


    /*
     *  Find the File Vector for a File Path
     */

static const struct FileDef *FindFile(const char *path)
{
    const struct FileDef *file;

    for (file = Files; file; file = file->Next)
	if (EqualStrings(path, file->Path))
	    break;
    return file;
}


#ifdef V36_COMPATIBILITY
    /*
     *	V36 (A3000 Beta ROM) compatible replacement for PeekQualifier()
     */

static u_short PeekQualifier36(void)
{
    u_short keys;

    /* Open Keyboard Device */

    if (!(KeyboardPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(KeyboardRequest = CreateIORequest(KeyboardPort,
					    sizeof(struct IOStdReq))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("keyboard.device", 0, (struct IORequest *)KeyboardRequest,
		   0))
	Alert(AN_LILO|AG_OpenDev|AO_KeyboardDev);
    KeyboardBase = (struct KeyboardBase *)KeyboardRequest->io_Device;

    KeyboardRequest->io_Command = KBD_READMATRIX;
    KeyboardRequest->io_Flags = IOF_QUICK;
    KeyboardRequest->io_Length = sizeof(KeyboardMatrix);
    KeyboardRequest->io_Data = KeyboardMatrix;
    DoIO((struct IORequest *)KeyboardRequest);

    keys = KeyboardMatrix[12];
    if (!(custom.potgor & 0x0100))
	keys |= IEQUALIFIER_MIDBUTTON;
    if (!(custom.potgor & 0x0400))
	keys |= IEQUALIFIER_RBUTTON;
    if (!(ciaa.pra & 0x40))
	keys |= IEQUALIFIER_LEFTBUTTON;

    /* Close Keyboard Device */

    CloseDevice((struct IORequest *)KeyboardRequest);
    DeleteIORequest(KeyboardRequest);
    DeleteMsgPort(KeyboardPort);

    return(keys);
}
#endif /* V36_COMPATIBILITY */


    /*
     *	Probe the Console
     */

static u_long ProbeConsole(void)
{
    u_long res;

#ifdef V36_COMPATIBILITY
    if (SysBase->Version >= 37)
#endif /* V36_COMPATIBILITY */
	res = PeekQualifier() ? TRUE : FALSE;
#ifdef V36_COMPATIBILITY
    else
	res = PeekQualifier36() ? TRUE : FALSE;
#endif /* V36_COMPATIBILITY */

    if (!res && Aux) {
	SetTimeOut(1);
	do {
	    if (SerialGetChar() != -1) {
		res = TRUE;
		while (SerialGetChar() != -1)
		    ;
		break;
	    }
	} while (!TimedOut());
	ClearTimeOut();
    }
    return(res);
}


    /*
     *	Replacement for exec.library/Debug()
     *
     *	We don't want to meet a debugger when DEL is pressed on the terminal
     */

asm(".text\n"
"	.align	4\n"
"NewDebug:\n"
"	rts"
);

extern void NewDebug(void);


    /*
     *	Open the Console
     */

static void OpenConsole(void)
{
    if (!ConsoleOpened) {
	IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", LIB_VERS);
	if (!IntuitionBase)
	    Alert(AN_LILO|AG_OpenLib|AO_Intuition);
	NewScreen.Width = GfxBase->NormalDisplayColumns;
	NewScreen.Height = GfxBase->NormalDisplayRows;
	NewScreen.DefaultTitle = Title;
	Screen = OpenScreen(&NewScreen);
	if (!Screen)
	    Alert(AN_LILO|AN_OpenScreen);
	NewWindow.Width = NewScreen.Width;
	NewWindow.Height = NewScreen.Height;
	NewWindow.Screen = Screen;
	NewWindow.Title = Title;
	Window = OpenWindow(&NewWindow);
	if (!Window)
	    Alert(AN_LILO|AN_OpenWindow);
	ConsolePort = CreateMsgPort();
	if (!ConsolePort)
	    Alert(AN_LILO|AG_NoSignal);
	ConsoleRequest = CreateIORequest(ConsolePort, sizeof(struct IOStdReq));
	if (!ConsoleRequest)
	    Alert(AN_LILO|AG_NoMemory);
	ConsoleRequest->io_Data = Window;
	if (OpenDevice("console.device", CONU_SNIPMAP,
		       (struct IORequest *)ConsoleRequest, CONFLAG_DEFAULT))
	    Alert(AN_LILO|AG_OpenDev|AO_ConsoleDev);
	ConsoleOpened = TRUE;
    }
}


    /*
     *	Close the Console
     */

static void CloseConsole()
{
    if (ConsoleOpened) {
	ConsoleOpened = FALSE;
	CloseDevice((struct IORequest *)ConsoleRequest);
	DeleteIORequest(ConsoleRequest);
	DeleteMsgPort(ConsolePort);
	CloseWindow(Window);
	CloseScreen(Screen);
	CloseLibrary((struct Library*)IntuitionBase);
    }
}


static int SerialGetChar(void)
{
    int ch;

    if (!(custom.intreqr & 0x0800))
	return -1;
    ch = custom.serdatr & 0xff;
    custom.intreq = 0x0800;
    return ch;
}

static void SerialPutChar(char c)
{
    if (c == '\n')
	SerialPutChar('\r');
    custom.serdat = (unsigned char)c | 0x100;
    while (!(custom.serdatr & 0x2000))
	;
}


    /*
     *	Print a String to the Console
     */

void Puts(const char *str)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = -1;
	ConsoleRequest->io_Data = (char *)str;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux) {
	    while (*str)
		SerialPutChar(*str++);
	}
	WaitIO((struct IORequest *)ConsoleRequest);
    }
#if 0
    else
	    while (*str) {
		if (*str == '\n')
		    asm volatile ("move.w %0,0xdff030; 9: btst #5,0xdff018; jeq 9b" : : "d" ('\r' | 0x100));
		asm volatile ("move.w %0,0xdff030; 9: btst #5,0xdff018; jeq 9b" : : "d" ((unsigned char)*str++ | 0x100));
	    }
#endif
}


    /*
     *	Put a Character to the Console
     */

void PutChar(char c)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux)
	    SerialPutChar(c);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
#if 0
    else
	    asm volatile ("move.w %0,0xdff030; 9: btst #5,0xdff018; jeq 9b" : : "d" ((unsigned char)c | 0x100));
#endif
}


    /*
     *	Get a Character from the Console
     */

static long GetChar(void)
{
    long res = -1;
    char c;

    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_READ;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	do
	    if (Aux && ((res = SerialGetChar()) != -1))
		break;
	    else if (CheckIO((struct IORequest *)ConsoleRequest)) {
		res = c;
		break;
	    }
	while (MasterMode || Debug || !TimedOut());
	AbortIO((struct IORequest *)ConsoleRequest);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
    return(res);
}


    /*
     *	Format and Print a String to the Console
     */

asm(".text\n"
"	.align	4\n"
"PutCharInBuf:\n"
"	move.l	%a0,-(%sp)\n"
"	lea	FmtBuffer+255(%pc),%a0\n"
"	cmp.l	%a0,%a3\n"
"	jne	1f\n"
"	clr.b	%a3@\n"
"	jra	2f\n"
"1:	move.b	%d0,(%a3)+\n"
"2:	move.l	(%sp)+,%a0\n"
"	rts\n"
);

extern void *PutCharInBuf(char c);

void Printf(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    RawDoFmt(fmt, args, (void (*)())PutCharInBuf, FmtBuffer);
    Puts(FmtBuffer);
    va_end(args);
}


    /*
     *	Read a Line from the Console
     */

static long ReadLine(u_long mode)
{
    long c;
    long length = 0;
    u_long eol = FALSE;

    while (!eol && ((c = GetChar()) != -1))
	switch (c) {
	    case '\r':
	    case '\n':
		InBuffer[length] = '\0';
		if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
		    PutChar('\n');
		eol = TRUE;
		break;
	    case '\b':
	    case 0x7f:
		if (length > 0) {
		    length--;
		    if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
			Puts("\b \b");
		}
		break;
	    case '\t':
		if (!MasterMode) {
		    strcpy(InBuffer, "help");
		    length = strlen("help");
		    switch (mode) {
			case ECHO_NORMAL:
			    Puts("help\n");
			    break;
			case ECHO_DOT:
			    Puts("++++\n");
			    break;
		    }
		    eol = TRUE;
		    break;
		}
	    default:
		if (length < sizeof(InBuffer) - 1) {
		    InBuffer[length++] = c;
		    switch (mode) {
			case ECHO_NORMAL:
			    PutChar(c);
			    break;
			case ECHO_DOT:
			    PutChar('+');
			    break;
		    }
		}
		break;
	}
    return(eol ? length : -1);
}


    /*
     *  Read the Command Line
     */

char *ReadCommandLine(void)
{
    ReadLine(ECHO_NORMAL);
    return(InBuffer);
}


    /*
     *	Set the Time Out Value
     */

static void SetTimeOut(u_long seconds)
{
    if (!TimeOutSignalMask) {
	TimerRequest->io_Command = TR_ADDREQUEST;
	TimerRequest->tv_secs = seconds;
	TimerRequest->tv_micro = 0;
	SendIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 1 << TimerPort->mp_SigBit;
    }
}


    /*
     *	Check for a Time Out
     */

static int TimedOut(void)
{
    int res = FALSE;

    if (TimeOutSignalMask && CheckIO((struct IORequest *)TimerRequest))
	res = TRUE;

    return(res);
}


    /*
     *	Clear Time Out
     */

static void ClearTimeOut(void)
{
    if (TimeOutSignalMask) {
	AbortIO((struct IORequest *)TimerRequest);
	WaitIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 0;
    }
}


    /*
     *	Sleep for a Specified Period
     */

static void Sleep(u_long micros)
{
    TimerRequest->io_Command = TR_ADDREQUEST;
    TimerRequest->io_Flags = IOF_QUICK;
    TimerRequest->tv_secs = micros / 1000000;
    TimerRequest->tv_micro = micros % 1000000;
    DoIO((struct IORequest *)TimerRequest);
}


    /*
     *  Check whether a Boot Image is available
     */

static int IsAvailable(const struct BootRecord *record)
{
    u_long type = record->OSType ? *record->OSType : BOS_LINUX;
    int res = TRUE;

    if (type != BOS_AMIGA) {
	if (!FindFile(record->Kernel))
	    res = FALSE;
	if (record->Ramdisk && !FindFile(record->Ramdisk))
	    res = FALSE;
    }
    return(res);
}


    /*
     *  List all available Boot Records
     */

void ListRecords(void)
{
    const struct BootRecord *record;

    Puts("\nBoot images:\n");
    for (record = BootRecords; record; record = record->Next) {
	if (record == DefaultBootRecord)
	    Puts("\e[1m");
	Printf("    %-31.31s (", record->Label);
	if (record->Alias)
	    Printf("`%s', ", record->Alias);
	switch (record->OSType ? *record->OSType : BOS_LINUX) {
	    case BOS_AMIGA:
		Puts("AmigaOS");
		break;
	    case BOS_LINUX:
		Puts("Linux/m68k");
		break;
	    case BOS_LINUXPPC:
		Puts("Linux/ppc");
		break;
	    case BOS_NETBSD:
		Puts("NetBSD-Amiga");
		break;
	    default:
		Puts("Unknown");
		break;
	}
	if (record->Password)
	    Puts(", Restricted");
	if (!IsAvailable(record))
	    Puts(", N/A");
	Puts(")\n");
	if (record == DefaultBootRecord)
	    Puts("\e[0m");
    }
}


    /*
     *  List all available Files
     */

void ListFiles(void)
{
    const struct FileDef *file;

    Puts("\nFiles:\n");
    for (file = Files; file; file = file->Next) {
	Printf("    %s (", file->Path);
	if (file->Vector)
	    Printf("%ld bytes", file->Vector->Size);
	else
	    Puts("N/A");
	Puts(")\n");
    }
}


    /*
     *	Ask the User for a Boot Operating System
     */

static void GetBootOS(void)
{
    const struct BootRecord *record;

    if (BootOptions->TimeOut ? *BootOptions->TimeOut : 20)
	SetTimeOut(*BootOptions->TimeOut);

    while (1) {
	PutChar('\n');
	Puts(Prompt);
	if ((ReadLine(ECHO_NORMAL)) < 1)
	    break;
	if (!strcmp(InBuffer, "help") || !strcmp(InBuffer, "?")) {
	    Printf("\n%s\nValid commands: su, help, ?\n", LiloVersion);
	    ListRecords();
	    PutChar('\n');
	} else if (!strcmp(InBuffer, "su")) {
	    if (BootOptions->MasterPassword) {
		Puts("\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		     strcmp(InBuffer, BootOptions->MasterPassword)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    MasterMode = TRUE;
	    Monitor();
	    break;
	} else {
	    if (!(record = FindBootRecord(InBuffer))) {
		Printf("\nNonexistent boot image `%s'.\n", InBuffer);
		continue;
	    }
	    if (!IsAvailable(record)) {
		Printf("\nBoot image `%s' is not available.\n", InBuffer);
		continue;
	    } else if (record->Password) {
		Puts("\n\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		    strcmp(InBuffer, record->Password)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    FillBootData(record, &BootData, FALSE);
	    break;
	}
    }
    ClearTimeOut();
}


    /*
     *  Fill the Boot Data according to a Boot Record
     */

void FillBootData(const struct BootRecord *record, struct BootData *boot,
		  int autoboot)
{
    int i;

    boot->OSType = record->OSType ? *record->OSType : BOS_LINUX;
    if (record->Kernel) {
	strncpy(boot->Kernel, record->Kernel, PATH_SIZE);
	boot->Kernel[PATH_SIZE - 1] = '\0';
    } else
	boot->Kernel[0] = '\0';
    if (record->Ramdisk) {
	strncpy(boot->Ramdisk, record->Ramdisk, PATH_SIZE);
	boot->Ramdisk[PATH_SIZE - 1] = '\0';
    } else
	boot->Ramdisk[0] = '\0';
    if (record->Args)
	strncpy(boot->Args, record->Args, CL_SIZE);
    else
	boot->Args[0] = '\0';
    strncat(boot->Args, " BOOT_IMAGE=", CL_SIZE);
    strncat(boot->Args, record->Label, CL_SIZE);
    if (autoboot)
	strncat(boot->Args, " auto", CL_SIZE);
    boot->Args[CL_SIZE - 1] = '\0';
    boot->Debug = Debug;
    boot->Model = record->Model ? *record->Model : AMI_UNKNOWN;
    boot->ChipSize = record->ChipSize ? *record->ChipSize : 0;
    boot->NumMemory = 0;
    for (i = 0; i < NUM_MEMINFO; i++) {
	if (record->FastChunks[i]) {
	    boot->FastChunks[i].Address = record->FastChunks[i][0];
	    boot->FastChunks[i].Size = record->FastChunks[i][1];
	    boot->NumMemory++;
	}
    }
    boot->Processor = record->Processor ? *record->Processor : 0;
}


    /*
     *  Boot the Linux/m68k Operating System
     */

static u_long BootLinux(const struct BootOptions *global,
			const struct BootData *boot, int ppc)
{
    struct linuxboot_args args;
    int processor, i, j;

    Printf("\nBooting Linux/%s...\n\n", ppc ? "ppc" : "m68k");

    memset(&args.bi, 0, sizeof(args.bi));
    processor = boot->Processor ? boot->Processor
				: global->Processor ? *global->Processor : 0;
    if (processor) {
	int cpu = processor / 100 % 10;
	int fpu = processor / 10 % 10;
	int mmu = processor % 10;
	if (cpu)
	    args.bi.cputype = 1 << (cpu - 1);
	if (fpu)
	    args.bi.fputype = 1 << (fpu - 1);
	if (mmu)
	    args.bi.mmutype = 1 << (mmu - 1);
    }
    if (boot->NumMemory) {
	for (i = 0, j = 0; i < boot->NumMemory; i++)
	    if (boot->FastChunks[i].Size) {
		args.bi.memory[j].addr = boot->FastChunks[i].Address;
		args.bi.memory[j].size = boot->FastChunks[i].Size;
		j++;
	    }
	if (j)
	    args.bi.num_memory = j;
    } else if (global->FastChunks[0])
	for (i = 0; global->FastChunks[i]; i++) {
	    args.bi.memory[i].addr = global->FastChunks[i][0];
	    args.bi.memory[i].size = global->FastChunks[i][1];
	}
    strncpy(args.bi.command_line, boot->Args, CL_SIZE);
    args.bi.command_line[CL_SIZE - 1] = '\0';
    if (boot->Model != AMI_UNKNOWN)
	args.bi.model = boot->Model;
    else if (global->Model)
	args.bi.model = *global->Model;
    if (boot->ChipSize)
	args.bi.chip_size = boot->ChipSize;
    else if (global->ChipSize)
	args.bi.chip_size = *global->ChipSize;

    args.kernelname = boot->Kernel[0] ? boot->Kernel : NULL;
    args.ramdiskname = boot->Ramdisk[0] ? boot->Ramdisk : NULL;
    args.debugflag = boot->Debug;
    args.keep_video = 0;
    args.reset_boards = 0;
    args.baud = BootOptions->Baud ? *BootOptions->Baud : 0;
    args.puts = Puts;
    args.getchar = GetChar;
    args.putchar = PutChar;
    args.printf = Printf;
    args.open = Open;
    args.seek = Seek;
    args.read = Read;
    args.close = Close;
    args.filesize = FileSize;
    args.sleep = Sleep;

    /* Do The Right Stuff */
    return(linuxboot(&args, ppc));
}


    /*
     *  Boot the NetBSD-Amiga Operating System
     */

static u_long BootNetBSD(struct BootData *boot)
{
    Puts("Not yet implemented :-)\n");
    return(FALSE);
}


    /*
     *	Functions needed by LinuxBoot
     *
     *	These functions are not complete implementations. They do only what's
     *	really necessary.
     */

#define BUFSIZE		65536

static int OpenFiles = 0;
static char *FileBuffer;
static const struct FileDef *CurrentFile;
static const struct Extent *CurrentExtent;
static u_int32_t FileBufferOffset, ExtentBufferOffset, BufferStart, SectorShift;
static struct MsgPort *DevicePort;
static struct IOStdReq *DeviceRequest;

static void FillNextBuffer(void)
{
    char *buf;
    int toread, size;
    u_int32_t max_sector;

    max_sector = 0xffffffffU >> SectorShift;
    buf = FileBuffer;
    toread = BUFSIZE >> SectorShift;
    while (toread) {
	size = toread;
	if (size > CurrentExtent->Count - ExtentBufferOffset)
	    size = CurrentExtent->Count - ExtentBufferOffset;

	if (!CurrentExtent->Block) {
	    memset(buf, 0, size << SectorShift);
	} else if ((CurrentExtent->Block + ExtentBufferOffset + size) <= max_sector) {
	    DeviceRequest->io_Command = CMD_READ;
	    DeviceRequest->io_Data = buf;
	    DeviceRequest->io_Length = size << SectorShift;
	    DeviceRequest->io_Offset = (CurrentExtent->Block + ExtentBufferOffset) << SectorShift;
	    if (DoIO((struct IORequest *)DeviceRequest))
		Alert(AN_LILO|AG_IOError);
	} else {
	    struct SCSICmd scsi_cmd;
	    char sense_data[4];
	    char read_cmd[10];

	    DeviceRequest->io_Command = HD_SCSICMD;
	    DeviceRequest->io_Data = &scsi_cmd;
	    DeviceRequest->io_Length = sizeof(struct SCSICmd);

	    scsi_cmd.scsi_Data = buf;
	    scsi_cmd.scsi_Length = size << SectorShift;
	    scsi_cmd.scsi_CmdLength = 10;
	    scsi_cmd.scsi_Flags = SCSIF_AUTOSENSE|SCSIF_READ;

	    scsi_cmd.scsi_Command = read_cmd;
	    scsi_cmd.scsi_SenseData = sense_data;
	    scsi_cmd.scsi_SenseLength = sizeof(sense_data);

	    read_cmd[0] = 0x28;
	    read_cmd[1] = 0;
	    *(u_int32_t *)(read_cmd + 2) = CurrentExtent->Block + ExtentBufferOffset;
	    read_cmd[6] = 0;
	    *(u_int16_t *)(read_cmd + 7) = size;
	    read_cmd[9] = 0;

	    if (DoIO((struct IORequest *)DeviceRequest))
		Alert(AN_LILO|AG_IOError);
	}

	buf += size << SectorShift;
	toread -= size;
	ExtentBufferOffset += size;
	if (ExtentBufferOffset == CurrentExtent->Count) {
	    if (BufferStart + (ExtentBufferOffset << SectorShift) >= CurrentFile->Extent->Size)
		break;
	    ExtentBufferOffset = 0;
	    CurrentExtent++;
	}
    }
}

static int OpenFile(const struct FileDef *file)
{
    if (OpenFiles) {
	Puts("Internal error: Only one file can be open\n");
	return -1;
    }

    if (!file->Extent)
	return -1;

    if (!FileBuffer) {
	FileBuffer = AllocMem(BUFSIZE, MEMF_PUBLIC|MEMF_CLEAR);
	if (!FileBuffer)
	    Alert(AN_LILO|AG_NoMemory);
    }

    if (file->device && file->device->device) {
	DevicePort = CreateMsgPort();
	if (!DevicePort)
	    Alert(AN_LILO|AG_NoSignal);
	DeviceRequest = CreateIORequest(DevicePort, sizeof(struct IOStdReq));
	if (!DeviceRequest)
	    Alert(AN_LILO|AG_NoMemory);
	if (OpenDevice(file->device->device, file->device->unit,
		       (struct IORequest *)DeviceRequest, 0))
	    Alert(AN_LILO|AG_OpenDev|AO_Unknown);
    } else
	DeviceRequest = BootRequest;

    CurrentFile = file;
    CurrentExtent = CurrentFile->Extent->Extent;
    SectorShift = CurrentFile->Extent->Shift;
    FileBufferOffset = ExtentBufferOffset = BufferStart = 0;
    FillNextBuffer();
    OpenFiles++;
    return 1;
}

static int Open(const char *path)
{
    const struct FileDef *file;

    file = FindFile(path);
    if (!file)
	return -1;

    return OpenFile(file);
}


static int Seek(int fd, int offset)
{
    int extent_offset, start;

    if (offset > CurrentFile->Extent->Size)
	return -1;

    FileBufferOffset = offset & (BUFSIZE - 1);
    if (offset >= BufferStart && offset < BufferStart + BUFSIZE)
	goto done;

    CurrentExtent = CurrentFile->Extent->Extent;
    BufferStart = offset & ~(BUFSIZE - 1);
    extent_offset = 0;
    start = BufferStart >> SectorShift;
    while (extent_offset + CurrentExtent->Count < start) {
	extent_offset += CurrentExtent->Count;
	CurrentExtent++;
    }
    ExtentBufferOffset = start - extent_offset;
    FillNextBuffer();
done:
    return 0;
}


static int Read(int fd, char *buf, int count)
{
    int size, toread;

    if (count > CurrentFile->Extent->Size - (BufferStart + FileBufferOffset))
	count = CurrentFile->Extent->Size - (BufferStart + FileBufferOffset);

    toread = count;
    while (toread) {
	size = toread;
	if (size > BUFSIZE - FileBufferOffset)
	    size = BUFSIZE - FileBufferOffset;

	memcpy(buf, FileBuffer + FileBufferOffset, size);
	buf += size;
	toread -= size;
	FileBufferOffset += size;
	if (FileBufferOffset == BUFSIZE) {
	    FillNextBuffer();
	    BufferStart += BUFSIZE;
	    FileBufferOffset = 0;
	}
    }
    return count;
}


static void Close(int fd)
{
    if (DevicePort) {
	CloseDevice((struct IORequest *)DeviceRequest);
	DeleteIORequest(DeviceRequest);
	DeleteMsgPort(DevicePort);
	DevicePort = NULL;
    }
    OpenFiles--;
}


static int FileSize(const char *path)
{
    const struct FileDef *file;

    file = FindFile(path);
    if (!file)
	return -1;
    return file->Extent->Size;
}


    /*
     *  Integer Math
     *
     *  Libnix uses utility.library if these aren't present
     */

asm("\n"
"		.globl	___umodsi3\n"
"		.globl	___udivsi3\n"
"		.globl	___udivsi4\n"
"		.globl	___mulsi3\n"
"\n"
"| D0.L = D0.L % D1.L unsigned\n"
"\n"
"___umodsi3:	movem.l	4(%sp),%d0/%d1\n"
"		jbsr	___udivsi4\n"
"		move.l	%d1,%d0\n"
"		rts\n"
"\n"
"| D0.L = D0.L / D1.L unsigned\n"
"\n"
"___udivsi3:	movem.l	4(%sp),%d0/%d1\n"
"___udivsi4:	move.l	%d3,-(%sp)\n"
"		move.l	%d2,-(%sp)\n"
"		move.l	%d1,%d3\n"
"		swap	%d1\n"
"		tstw	%d1\n"
"		jne	1f\n"
"		move.w	%d0,%d2\n"
"		clr.w	%d0\n"
"		swap	%d0\n"
"		divu	%d3,%d0\n"
"		move.l	%d0,%d1\n"
"		swap	%d0\n"
"		move.w	%d2,%d1\n"
"		divu	%d3,%d1\n"
"		move.w	%d1,%d0\n"
"		clr.w	%d1\n"
"		swap	%d1\n"
"		jra	4f\n"
"1:		move.l	%d0,%d1\n"
"		swap	%d0\n"
"		clr.w	%d0\n"
"		clr.w	%d1\n"
"		swap	%d1\n"
"		moveq	#16-1,%d2\n"
"2:		add.l	%d0,%d0\n"
"		addx.l	%d1,%d1\n"
"		cmp.l	%d1,%d3\n"
"		jhi	3f\n"
"		sub.l	%d3,%d1\n"
"		addq.w	#1,%d0\n"
"3:		dbra	%d2,2b\n"
"4:		move.l	(%sp)+,%d2\n"
"		move.l	(%sp)+,%d3\n"
"		rts\n"
"\n"
"| D0 = D0 * D1\n"
"\n"
"___mulsi3:	movem.l	4(%sp),%d0/%d1\n"
"		move.l	%d3,-(%sp)\n"
"		move.l	%d2,-(%sp)\n"
"		move.w	%d1,%d2\n"
"		mulu	%d0,%d2\n"
"		move.l	%d1,%d3\n"
"		swap	%d3\n"
"		mulu	%d0,%d3\n"
"		swap	%d3\n"
"		clr.w	%d3\n"
"		add.l	%d3,%d2\n"
"		swap	%d0\n"
"		mulu	%d1,%d0\n"
"		swap	%d0\n"
"		clr.w	%d0\n"
"		add.l	%d2,%d0\n"
"		move.l	(%sp)+,%d2\n"
"		move.l	(%sp)+,%d3\n"
"		rts\n"
);
