/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 */


    /*
     *  Boot Parameters
     */

struct BootData {
    u_long OSType;
    char Kernel[PATH_SIZE];
    char Ramdisk[PATH_SIZE];
    char Args[CL_SIZE];
    u_long Debug;
    u_long Model;
    u_long ChipSize;
    u_int NumMemory;
    struct FastChunk {
	u_long Address;
	u_long Size;
    } FastChunks[NUM_MEMINFO];
    u_long Processor;
};


    /*
     *	Exec Library Definitions
     */

#define AT_DeadEnd	(0x80000000)
#define AG_NoMemory	(0x00010000)
#define AG_NoSignal	(0x00070000)
#define AG_OpenLib	(0x00030000)
#define AG_OpenDev	(0x00040000)
#define AG_OpenRes	(0x00050000)
#define AG_IOError	(0x00060000)
#define AO_GraphicsLib	(0x00008002)
#define AO_Intuition	(0x00008004)
#define AO_ExpansionLib	(0x0000800A)
#define AO_ConsoleDev	(0x00008011)
#define AO_KeyboardDev	(0x00008013)
#define AO_TimerDev	(0x00008015)
#define AO_Unknown	(0x00008035)
#define AN_OpenScreen	(0x84010007)
#define AN_OpenWindow	(0x8401000B)

struct MsgPort {
	u_char fill1[15];
	u_char mp_SigBit;
	u_char fill2[18];
};

struct Message {
	u_char fill1[14];
	struct MsgPort *mn_ReplyPort;
	u_int16_t mn_Length;
};

struct IOStdReq {
	struct Message io_Message;
	struct Device *io_Device;
	u_char fill2[4];
	u_int16_t io_Command;
	u_int8_t io_Flags;
	int8_t io_Error;
	u_int32_t io_Actual;
	u_int32_t io_Length;
	void *io_Data;
	u_int32_t io_Offset;
};

#define CMD_READ	(2)
#define CMD_WRITE	(3)
#define KBD_READMATRIX	(10)
#define HD_SCSICMD	(28)

#define IOF_QUICK	(1<<0)

struct timerequest {
	struct Message io_Message;
	u_char fill1[8];
	u_int16_t io_Command;
	u_int8_t io_Flags;
	int8_t io_Error;
	u_int32_t tv_secs;
	u_int32_t tv_micro;
};

#define UNIT_VBLANK	(1)
#define TR_ADDREQUEST	(9)


#define CONU_SNIPMAP	(3)
#define CONFLAG_DEFAULT	(0)

#define IEQUALIFIER_LSHIFT	(0x0001)
#define IEQUALIFIER_RSHIFT	(0x0002)
#define IEQUALIFIER_CAPSLOCK	(0x0004)
#define IEQUALIFIER_CONTROL	(0x0008)
#define IEQUALIFIER_LALT	(0x0010)
#define IEQUALIFIER_RALT	(0x0020)
#define IEQUALIFIER_LCOMMAND	(0x0040)
#define IEQUALIFIER_RCOMMAND	(0x0080)
#define IEQUALIFIER_MIDBUTTON	(0x1000)
#define IEQUALIFIER_RBUTTON	(0x2000)
#define IEQUALIFIER_LEFTBUTTON	(0x4000)


struct SCSICmd {
	void *scsi_Data;
	u_int32_t scsi_Length;
	u_int32_t scsi_Actual;
	void *scsi_Command;
	u_int16_t scsi_CmdLength;
	u_int16_t scsi_CmdActual;
	u_int8_t scsi_Flags;
	u_int8_t scsi_Status;
	void *scsi_SenseData;
	u_int16_t scsi_SenseLength;
	u_int16_t scsi_SenseActual;
};

#define SCSIF_WRITE	0
#define SCSIF_READ	1
#define SCSIF_AUTOSENSE	2


struct IORequest;
struct Library;


    /*
     *	Intuition Library Definitions
     */

struct NewScreen {
    short LeftEdge;
    short TopEdge;
    short Width;
    short Height;
    short Depth;
    u_char DetailPen;
    u_char BlockPen;
    u_short ViewModes;
    u_short Type;
    struct TextAttr *Font;
    u_char *DefaultTitle;
    struct Gadget *Gadgets;
    struct BitMap *CustomBitMap;
};

#define CUSTOMSCREEN	(0x000f)

struct NewWindow {
    short LeftEdge;
    short TopEdge;
    short Width;
    short Height;
    u_char DetailPen;
    u_char BlockPen;
    u_long IDCMPFlags;
    u_long Flags;
    struct Gadget *FirstGadget;
    struct Image *CheckMark;
    u_char *Title;
    struct Screen *Screen;
    struct BitMap *BitMap;
    short MinWidth;
    short MinHeight;
    u_short MaxWidth;
    u_short MaxHeight;
    u_short Type;
};

#define WFLG_SIMPLE_REFRESH	(0x00000040)
#define WFLG_ACTIVATE		(0x00001000)

struct Window;


    /*
     *	Amiga Shared Library/Device Functions
     */

extern const struct ExecBase *SysBase;

#define LVOAbortIO		(-0x1e0)
#define LVOAlert		(-0x6c)
#define LVOCheckIO		(-0x1d4)
#define LVOCloseDevice		(-0x1c2)
#define LVOCloseLibrary		(-0x19e)
#define LVOCreateIORequest	(-0x28e)
#define LVOCreateMsgPort	(-0x29a)
#define LVODebug		(-0x72)
#define LVODeleteIORequest	(-0x294)
#define LVODeleteMsgPort	(-0x2a0)
#define LVODoIO			(-0x1c8)
#define LVOOpenDevice		(-0x1bc)
#define LVOOpenLibrary		(-0x228)
#define LVORawDoFmt		(-0x20a)
#define LVORawMayGetChar	(-0x1fe)
#define LVORawPutChar		(-0x204)
#define LVOSendIO		(-0x1ce)
#define LVOSetFunction		(-0x1a4)
#define LVOWaitIO		(-0x1da)

#define BASE(val)	register u_long base asm("d6") = (u_long)(val)
#define RES(type)	register type res asm("d0")
#define REG(reg, val)	register u_long reg asm(#reg) = (u_long)(val)

static inline char AbortIO(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);
	RES(char);

	asm volatile ("exg %%d6,%%a6; jsr -0x1e0(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a1)
		      : "r" (base), "1" (a1)
		      : "a0", "d1", "memory");
	return res;
}

static inline void Alert(u_long alertNum)
{
	BASE(SysBase);
	REG(d7, alertNum);

	asm volatile ("exg %%d6,%%a6; jsr -0x6c(%%a6); exg %%d6,%%a6"
		      : /* no output */
		      : "r" (base), "r" (d7)
		      : "a0", "a1", "d0", "d1", "memory");
}

static inline u_short CheckIO(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);
	RES(u_short);

	asm volatile ("exg %%d6,%%a6; jsr -0x1d4(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a1)
		      : "r" (base), "1" (a1)
		      : "a0", "d1", "memory");
	return res;
}

static inline void CloseDevice(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);

	asm volatile ("exg %%d6,%%a6; jsr -0x1c2(%%a6); exg %%d6,%%a6"
		      : "=r" (a1)
		      : "r" (base), "0" (a1)
		      : "a0", "d0", "d1", "memory");
}

static inline void CloseLibrary(struct Library *library)
{
	BASE(SysBase);
	REG(a1, library);

	asm volatile ("exg %%d6,%%a6; jsr -0x19e(%%a6); exg %%d6,%%a6"
		      : "=r" (a1)
		      : "r" (base), "0" (a1)
		      : "a0", "d0", "d1", "memory");
}

static inline void *CreateIORequest(struct MsgPort *port, u_long size)
{
	BASE(SysBase);
	REG(a0, port);
	REG(d0, size);
	RES(void *);

	asm volatile ("exg %%d6,%%a6; jsr -0x28e(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0)
		      : "r" (base), "0" (d0), "1" (a0)
		      : "a1", "d1", "memory");
	return res;
}

static inline struct MsgPort *CreateMsgPort(void)
{
	BASE(SysBase);
	RES(struct MsgPort *);

	asm volatile ("exg %%d6,%%a6; jsr -0x29a(%%a6); exg %%d6,%%a6"
		      : "=r" (res)
		      : "r" (base)
		      : "a0", "a1", "d1", "memory");
	return res;
}

static inline void DeleteIORequest(void *ioRequest)
{
	BASE(SysBase);
	REG(a0, ioRequest);

	asm volatile ("exg %%d6,%%a6; jsr -0x294(%%a6); exg %%d6,%%a6"
		      : "=r" (a0)
		      : "r" (base), "0" (a0)
		      : "a1", "d0", "d1", "memory");
}

static inline void DeleteMsgPort(struct MsgPort *port)
{
	BASE(SysBase);
	REG(a0, port);

	asm volatile ("exg %%d6,%%a6; jsr -0x2a0(%%a6); exg %%d6,%%a6"
		      : "=r" (a0)
		      : "r" (base), "0" (a0)
		      : "a1", "d0", "d1", "memory");
}

static inline char DoIO(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);
	RES(char);

	asm volatile ("exg %%d6,%%a6; jsr -0x1c8(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a1)
		      : "r" (base), "1" (a1)
		      : "a0", "d1", "memory");
	return res;
}

static inline char OpenDevice(const u_char *devName, u_long unit,
			      struct IORequest *ioRequest, u_long flags)
{
	BASE(SysBase);
	REG(a0, devName);
	REG(d0, unit);
	REG(a1, ioRequest);
	REG(d1, flags);
	RES(char);

	asm volatile ("exg %%d6,%%a6; jsr -0x1bc(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (d1), "=r" (a0), "=r" (a1)
		      : "r" (base), "0" (d0), "1" (d1), "2" (a0), "3" (a1)
		      : "memory");
	return res;
}

static inline struct Library *OpenLibrary(const u_char *libName, u_long version)
{
	BASE(SysBase);
	REG(a1, libName);
	REG(d0, version);
	RES(struct Library *);

	asm volatile ("exg %%d6,%%a6; jsr -0x228(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a1)
		      : "r" (base), "0" (d0), "1" (a1)
		      : "d1", "a0", "memory");
	return res;
}

static inline void *RawDoFmt(const u_char *formatString, void *dataStream,
			     void (*putChProc)(), void *putChData)
{
	BASE(SysBase);
	REG(a0, formatString);
	REG(a1, dataStream);
	REG(a2, putChProc);
	REG(a3, putChData);
	RES(void *);

	asm volatile ("exg %%d6,%%a6; jsr -0x20a(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0), "=r" (a1)
		      : "r" (base), "1" (a0), "2" (a1), "r" (a2), "r" (a3)
		      : "d1", "memory");
	return res;
}

static inline long RawMayGetChar(void)
{
	BASE(SysBase);
	RES(long);

	asm volatile ("exg %%d6,%%a6; jsr -0x1fe(%%a6); exg %%d6,%%a6"
		      : "=r" (res)
		      : "r" (base)
		      : "a0", "a1", "d1", "memory");
	return res;
}

static inline void RawPutChar(char c)
{
	BASE(SysBase);
	REG(d0, c);

	asm volatile ("exg %%d6,%%a6; jsr -0x204(%%a6); exg %%d6,%%a6"
		      : "=r" (d0)
		      : "r" (base), "0" (d0)
		      : "a0", "a1", "d1", "memory");
}

static inline void SendIO(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);

	asm volatile ("exg %%d6,%%a6; jsr -0x1ce(%%a6); exg %%d6,%%a6"
		      : "=r" (a1)
		      : "r" (base), "0" (a1)
		      : "a0", "d0", "d1", "memory");
}

static inline void *SetFunction(struct Library *library, long funcOffset,
				  void *funcEntry)
{
	BASE(SysBase);
	REG(a1, library);
	REG(a0, funcOffset);
	REG(d0, funcEntry);
	RES(void *);

	asm volatile ("exg %%d6,%%a6; jsr -0x1a4(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0), "=r" (a1)
		      : "r" (base), "0" (d0), "1" (a0), "2" (a1)
		      : "d1", "memory");
	return res;
}

static inline char WaitIO(struct IORequest *ioRequest)
{
	BASE(SysBase);
	REG(a1, ioRequest);
	RES(char);

	asm volatile ("exg %%d6,%%a6; jsr -0x1da(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a1)
		      : "r" (base), "1" (a1)
		      : "a0", "d1", "memory");
	return res;
}


extern const struct InputBase *InputBase;

#define LVOPeekQualifier	(-0x2a)

static inline u_short PeekQualifier(void)
{
	BASE(InputBase);
	RES(u_short);

	asm volatile ("exg %%d6,%%a6; jsr -0x2a(%%a6); exg %%d6,%%a6"
		      : "=r" (res)
		      : "r" (base)
		      : "a0", "a1", "d1", "memory");
	return res;
}


extern const struct IntuitionBase *IntuitionBase;

#define LVOCloseScreen		(-0x42)
#define LVOCloseWindow		(-0x48)
#define LVOOpenScreen		(-0xc6)
#define LVOOpenWindow		(-0xcc)

static inline u_short CloseScreen(struct Screen *screen)
{
	BASE(IntuitionBase);
	REG(a0, screen);
	RES(u_short);

	asm volatile ("exg %%d6,%%a6; jsr -0x42(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0)
		      : "r" (base), "1" (a0)
		      : "a1", "d1", "memory");
	return res;
}

static inline void CloseWindow(struct Window *window)
{
	BASE(IntuitionBase);
	REG(a0, window);

	asm volatile ("exg %%d6,%%a6; jsr -0x48(%%a6); exg %%d6,%%a6"
		      : "=r" (a0)
		      : "r" (base), "0" (a0)
		      : "a1", "d0", "d1", "memory");
}

static inline struct Screen *OpenScreen(struct NewScreen *newScreen)
{
	BASE(IntuitionBase);
	REG(a0, newScreen);
	RES(struct Screen *);

	asm volatile ("exg %%d6,%%a6; jsr -0xc6(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0)
		      : "r" (base), "1" (a0)
		      : "a1", "d1", "memory");
	return res;
}

static inline struct Window *OpenWindow(struct NewWindow *newWindow)
{
	BASE(IntuitionBase);
	REG(a0, newWindow);
	RES(struct Window *);

	asm volatile ("exg %%d6,%%a6; jsr -0xcc(%%a6); exg %%d6,%%a6"
		      : "=r" (res), "=r" (a0)
		      : "r" (base), "1" (a0)
		      : "a1", "d1", "memory");
	return res;
}

#undef BASE
#undef RES
#undef REG
